/**************************************************************************/
/*                                                                        */
/*  This file is part of Frama-Clang                                      */
/*                                                                        */
/*  Copyright (C) 2012-2017                                               */
/*    CEA (Commissariat à l'énergie atomique et aux énergies              */
/*         alternatives)                                                  */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 2.1.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version 2.1                 */
/*  for more details (enclosed in the file LICENSE).                      */
/*                                                                        */
/**************************************************************************/

//
// Description:
//   Implementation of ACSL++ tokens.
//

#include "ACSLToken.h"

namespace Acsl { namespace DLexer {

void
AbstractToken::_write(std::ostream& out, PersistentFormat& format) const {
  if (!format.isPretty()) {
    unsigned int params = queryParams();
    out.write((const char*) &params, sizeof(unsigned));
  }
  else {
    switch (queryTypeField()) {
      case TUndefined: out << "undefined"; break;
      case TIdentifier: out << "identifier"; break;
      case TKeyword: out << "keyword"; break;
      case TLiteral: out << "string_literal"; break;
      case TOperatorPunctuator: out << "operator"; break;
      case TComment: out << "comment"; break;
      default: break;
    };
  };
}

void
CommentToken::_write(std::ostream& out, PersistentFormat& format) const {
  inherited::_write(out, format);
  if (!format.isPretty()) {
    unsigned size = _content.length();
    out.write((const char*) &size, sizeof(unsigned));
    out.write(_content.c_str(), size);
  }
  else
    out << " -> " << _content;
}

void
IdentifierToken::_write(std::ostream& out, PersistentFormat& format) const {
  inherited::_write(out, format);
  if (!format.isPretty()) {
    unsigned size = _content.length();
    out.write((const char*) &size, sizeof(unsigned));
    out.write(_content.c_str(), size);
  }
  else
    out << " -> " << _content;
}


std::pair<std::string, KeywordToken::Type>
KeywordToken::mapUnprotected[] = {
  std::make_pair("auto", TAuto),
  std::make_pair("bool", TBool),
  std::make_pair("case", TCase),
  std::make_pair("char", TChar),
  std::make_pair("char16_t", TChar16),
  std::make_pair("char32_t", TChar32),
  std::make_pair("class", TClass),
  std::make_pair("const", TConst),
  std::make_pair("const_cast", TConstCast),
  std::make_pair("double", TDouble),
  std::make_pair("dynamic_cast", TDynamicCast),
  std::make_pair("else", TElse),
  std::make_pair("enum", TEnum),
  std::make_pair("false", TFalse),
  std::make_pair("float", TFloat),
  std::make_pair("for", TFor),
  std::make_pair("global", TGlobal),
  std::make_pair("if", TIf),
  std::make_pair("int", TInt),
  std::make_pair("long", TLong),
  std::make_pair("loop", TLoop),
  std::make_pair("operator", TOperator),
  std::make_pair("reinterpret_cast", TReinterpretCast),
  std::make_pair("short", TShort),
  std::make_pair("signed", TSigned),
  std::make_pair("sizeof", TSizeof),
  std::make_pair("static", TStatic),
  std::make_pair("static_cast", TStaticCast),
  std::make_pair("struct", TStruct),
  std::make_pair("true", TTrue),
  std::make_pair("typeId", TTypeId),
  std::make_pair("typename", TTypename),
  std::make_pair("union", TUnion),
  std::make_pair("unsigned", TUnsigned),
  std::make_pair("virtual", TVirtual),
  std::make_pair("void", TVoid),
  std::make_pair("volatile", TVolatile),
  std::make_pair("wchar_t", TWcharT),
  std::make_pair("while", TWhile),
  std::make_pair("integer", TInteger),
  std::make_pair("invariant", TInvariant),
  std::make_pair("real", TReal),
  std::make_pair("boolean", TBoolean),
  std::make_pair("logic", TLogic),
  std::make_pair("predicate", TPredicate),
  std::make_pair("lemma", TLemma),
  std::make_pair("axiomatic", TAxiomatic),
  std::make_pair("axiom", TAxiom),
  std::make_pair("ensures", TEnsures),
  std::make_pair("assumes", TAssumes),
  std::make_pair("requires", TRequires),
  std::make_pair("terminates", TTerminates),
  std::make_pair("decreases", TDecreases),
  std::make_pair("assert", TAssert),
  std::make_pair("assigns", TAssigns),
  std::make_pair("reads", TReads),
  std::make_pair("writes", TWrites),
  std::make_pair("behavior", TBehavior),
  std::make_pair("complete", TComplete),
  std::make_pair("disjoint", TDisjoint),
  std::make_pair("behaviors", TBehaviors),
  std::make_pair("type", TType),
  std::make_pair("inductive", TInductive),
  std::make_pair("exits", TExits),
  std::make_pair("breaks", TBreaks),
  std::make_pair("continues", TContinues),
  std::make_pair("returns", TReturns),
  std::make_pair("variant", TVariant),
  std::make_pair("frees", TFrees),
  std::make_pair("allocates", TAllocates)
};

std::tr1::unordered_map<std::string, KeywordToken::Type>
KeywordToken::_unprotectedKeywords(
  mapUnprotected,
  mapUnprotected + sizeof(mapUnprotected) / sizeof(mapUnprotected[0]));

std::pair<std::string, KeywordToken::Type>
KeywordToken::mapProtected[] = {
  std::make_pair("forall", TForall),
  std::make_pair("exists", TExists),
  std::make_pair("true", TTrue),
  std::make_pair("false", TFalse),
  std::make_pair("let", TLet),
  std::make_pair("nothing", TNothing),
  std::make_pair("at", TAt),
  std::make_pair("old", TOld),
  std::make_pair("result", TResult),
  std::make_pair("match", TMatch),
  std::make_pair("lambda", TLambda),
  std::make_pair("sum", TSum),
  std::make_pair("product", TProduct),
  std::make_pair("length", TLength),
  std::make_pair("block_length", TBlockLength),
  std::make_pair("offset", TOffset),
  std::make_pair("t_iso", TTIso),
  std::make_pair("origin", TOrigin),
  std::make_pair("object", TObject),
  std::make_pair("null", TNull),
  std::make_pair("base_addr", TBaseAddr),
  std::make_pair("allocation", TAllocation),
  std::make_pair("allocable", TAllocable),
  std::make_pair("freeable", TFreeable),
  std::make_pair("fresh", TFresh),
  std::make_pair("separated", TSeparated),
  std::make_pair("frees", TFrees),
  std::make_pair("allocates", TAllocates),
  std::make_pair("allocation_status", TAllocationStatus),
  std::make_pair("static", TLStatic),
  std::make_pair("register", TLRegister),
  std::make_pair("automatic", TLAutomatic),
  std::make_pair("dynamic", TLDynamic),
  std::make_pair("unallocated", TLUnallocated),
  std::make_pair("initialized", TInitialized),
  std::make_pair("specified", TSpecified),
  std::make_pair("with", TWith),
  std::make_pair("num_of", TNumOf),
  std::make_pair("union", TLUnion),
  std::make_pair("inter", TLInter),
  std::make_pair("subset", TSubset),
  std::make_pair("empty", TEmpty),
  std::make_pair("valid", TValid),
  std::make_pair("valid_read", TValidRead),
  std::make_pair("valid_index", TValidIndex),
  std::make_pair("valid_range", TValidRange),
  std::make_pair("set", TSet),
  std::make_pair("exit_status", TExitStatus),
  std::make_pair("from", TFrom),
  std::make_pair("weak", TWeak),
  std::make_pair("strong", TStrong),
  std::make_pair("ghost", TGhost),
  //  std::make_pair("up", TUp),
  //std::make_pair("down", TDown),
  //std::make_pair("nearest_away", TNearestAway),
  //std::make_pair("nearest_even", TNearestEven),
  //std::make_pair("round_float", TRoundFloat),
  //std::make_pair("round_double", TRoundDouble),
  //std::make_pair("is_finite", TIsFinite),
  //std::make_pair("is_NaN", TIsNaN),
  //std::make_pair("min", TMin),
  //std::make_pair("max", TMax),
  //std::make_pair("abs", TAbs),
  //std::make_pair("sqrt", TSqrt),
  //std::make_pair("pow", TPow),
  //std::make_pair("ceil", TCeil),
  //std::make_pair("floor", TFloor),
  //std::make_pair("exp", TExp),
  //std::make_pair("log", TLog),
  //std::make_pair("log10", TLog10),
  //std::make_pair("cos", TCos),
  //std::make_pair("sin", TSin),
  //std::make_pair("tan", TTan),
  //std::make_pair("cosh", TCosh),
  //std::make_pair("sinh", TSinh),
  //std::make_pair("tanh", TTanh),
  //std::make_pair("acos", TAcos),
  //std::make_pair("asin", TAsin),
  //std::make_pair("atan", TAtan),
  //std::make_pair("atan2", TAtan2),
  //std::make_pair("hypot", THypot),
  //std::make_pair("exact", TExact),
  //std::make_pair("round_error", TRoundError),
};

/* =
  { { "asm", TAsm },
    { "auto", TAuto },
    { "bool", TBool },
    { "break", TBreak },
    { "case", TCase },
    { "catch", TCatch },
    { "class", TClass },
    { "char", TChar },
    { "class", TClass },
    { "const", TConst },
    { "const_cast", TConstCast },
    { "continue", TContinue },
    { "default", TDefault },
    { "delete", TDelete },
    { "do", TDo },
    { "double", TDouble },
    { "dynamic_cast", TDynamicCast },
    { "else", TElse },
    { "enum", TEnum },
    { "explicit", TExplicit },
    { "export", TExport },
    { "extern", TExtern },
    { "false", TFalse },
    { "float", TFloat },
    { "for", TFor },
    { "friend", TFriend },
    { "global", TGlobal },
    { "goto", TGoto },
    { "if", TIf },
    { "inline", TInline },
    { "int", TInt },
    { "invariant", TInvariant },
    { "long", TLong },
    { "mutable", TMutable },
    { "namespace", TNamespace },
    { "new", TNew },
    { "operator", TOperator },
    { "private", TPrivate },
    { "protected", TProtected },
    { "public", TPublic },
    { "register", TRegister },
    { "reinterpret_cast", TReinterpretCast },
    { "return", TReturn },
    { "short", TShort },
    { "signed", TSigned },
    { "sizeof", TSizeof },
    { "static", TStatic },
    { "staticCast", TStaticCast },
    { "struct", TStruct },
    { "switch", TSwitch },
    { "template", TTemplate },
    { "throw", TThrow },
    { "true", TTrue },
    { "try", TTry },
    { "typedef", TTypedef },
    { "typeId", TTypeId },
    { "typename", TTypename },
    { "union", TUnion },
    { "unsigned", TUnsigned },
    { "using", TUsing },
    { "variant", TVariant },
    { "virtual", TVirtual },
    { "void", TVoid },
    { "volatile", TVolatile },
    { "wchar_t", TWcharT },
    { "while", TWhile },
    { "integer", TInteger },
    { "real", TReal },
    { "boolean", TBoolean },
    { "logic", TLogic },
    { "predicate", TPredicate },
    { "lemma", TLemma },
    { "axiom", TAxiom },
    { "axiomatic", TAxiomatic },
    { "ensures", TEnsures },
    { "assumes", TAssumes },
    { "requires", TRequires },
    { "terminates", TTerminates },
    { "decreases", TDecreases },
    { "assert", TAssert },
    { "assigns", TAssigns },
    { "reads", TReads },
    { "writes", TWrites },
    { "behavior", TBehavior },
    { "complete", TComplete },
    { "disjoint", TDisjoint },
    { "behaviors", TBehaviors },
    { "type", TType },
    { "inductive", TInductive },
    { "exists", TExists },
    { "breaks", TBreaks },
    { "continues", TContinues },
    { "returns", TReturns },
  };
*/

std::tr1::unordered_map<std::string, KeywordToken::Type>
KeywordToken::_protectedKeywords(
  mapProtected,
  mapProtected + sizeof(mapProtected) / sizeof(mapProtected[0]));

/*
std::tr1::unordered_map<std::string, KeywordToken::Type>
KeywordToken::_protectedKeywords =
  { { "forall", TForall },
    { "exists", TLExists },
    { "true", TTrue },
    { "false", TFalse },
    { "let", TLet },
    { "nothing", TNothing },
    { "at", TAt },
    { "old", TOld },
    { "result", TResult },
    { "match", TMatch },
    { "lambda", TLambda },
    { "sum", TSum },
    { "product", TProduct },
    { "length", TLength },
    { "block_length", TBlockLength },
    { "offset", TOffset },
    { "t_iso", TTIso },
    { "origin", TOrigin },
    { "object", TObject },
    { "null", TNull },
    { "baseAddr", TBaseAddr },
    { "allocation", TAllocation },
    { "allocable", TAllocable },
    { "freeable", TFreeable },
    { "fresh", TFresh },
    { "separated", TSeparated },
    { "frees", TFrees },
    { "allocates", TAllocates },
    { "allocation_status", TAllocationStatus },
    { "static", TLStatic },
    { "register", TLRegister },
    { "automatic", TLAutomatic },
    { "dynamic", TLDynamic },
    { "unallocated", TLUnallocated },
    { "initialized", TInitialized },
    { "specified", TSpecified },
    { "with", TWith },
    { "numOf", TNumOf },
    { "union", TLUnion },
    { "inter", TLInter },
    { "subset", TSubset },
    { "empty", TEmpty },
    { "valid", TValid },
    { "valid_read", TValidRead },
    { "valid_index", TValidIndex },
    { "valid_range", TValidRange },
    { "loop", TLoop },
    { "set", TSet },
    { "exitStatus", TExitStatus },
    { "from", TFrom },
    { "weak", TWeak },
    { "strong", TStrong },
    { "ghost", TGhost },
    { "up", TUp },
    { "down", TDown },
    { "nearest_away", TNearestAway },
    { "nearest_even", TNearestEven },
    { "round_float", TRoundFloat },
    { "round_double", TRoundDouble },
    { "is_finite", TIsFinite },
    { "is_NaN", TIsNaN },
    { "min", TMin },
    { "max", TMax },
    { "abs", TAbs },
    { "sqrt", TSqrt },
    { "pow", TPow },
    { "ceil", TCeil },
    { "floor", TFloor },
    { "exp", TExp },
    { "log", TLog },
    { "log10", TLog10 },
    { "cos", TCos },
    { "sin", TSin },
    { "tan", TTan },
    { "cosh", TCosh },
    { "sinh", TSinh },
    { "tanh", TTanh },
    { "acos", TAcos },
    { "asin", TAsin },
    { "atan", TAtan },
    { "atan2", TAtan2 },
    { "hypot", THypot },
    { "exact", TExact },
    { "round_error", TRoundError },
  };
*/

void
KeywordToken::_write(std::ostream& out, PersistentFormat& format) const {
  inherited::_write(out, format);
  if (format.isPretty()) {
    out << " -> ";
    switch (queryOwnField()) {
      case TAuto: out << "auto"; break;
      case TBool: out << "bool"; break;
      case TCase: out << "case"; break;
      case TChar: out << "char"; break;
      case TClass: out << "class"; break;
      case TConst: out << "const"; break;
      case TConstCast: out << "const_cast"; break;
      case TDouble: out << "double"; break;
      case TDynamicCast: out << "dynamic_cast"; break;
      case TElse: out << "else"; break;
      case TEnum: out << "enum"; break;
      case TFalse: out << "false"; break;
      case TFloat: out << "float"; break;
      case TFor: out << "for"; break;
      case TIf: out << "if"; break;
      case TInt: out << "int"; break;
      case TLong: out << "long"; break;
      case TOperator: out << "operator"; break;
      case TReinterpretCast: out << "reinterpret_cast"; break;
      case TShort: out << "short"; break;
      case TSigned: out << "signed"; break;
      case TSizeof: out << "sizeof"; break;
      case TStaticCast: out << "static_cast"; break;
      case TStruct: out << "struct"; break;
      case TTrue: out << "true"; break;
      case TTypeId: out << "typeid"; break;
      case TTypename: out << "typename"; break;
      case TUnion: out << "union"; break;
      case TUnsigned: out << "unsigned"; break;
      case TVirtual: out << "virtual"; break;
      case TVoid: out << "void"; break;
      case TVolatile: out << "volatile"; break;
      case TWcharT: out << "wchar_t"; break;
      case TWhile: out << "while"; break;

      case TForall: out << "\\forall"; break;
      case TExists: out << "\\exists"; break;
      case TInteger: out << "integer"; break;
      case TReal: out << "real"; break;
      case TBoolean: out << "boolean"; break;
      case TLTrue: out << "\\true"; break;
      case TLFalse: out << "\\false"; break;
      case TLet: out << "\\let"; break;
      case TLogic: out << "logic"; break;
      case TPredicate: out << "predicate"; break;
      case TLemma: out << "lemma"; break;
      case TAxiomatic: out << "axiomatic"; break;
      case TAxiom: out << "axiom"; break;
      case TEnsures: out << "ensures"; break;
      case TAssumes: out << "assumes"; break;
      case TRequires: out << "requires"; break;
      case TTerminates: out << "terminates"; break;
      case TDecreases: out << "decreases"; break;
      case TAssert: out << "assert"; break;
      case TAssigns: out << "assigns"; break;
      case TReads: out << "reads"; break;
      case TWrites: out << "writes"; break;
      case TNothing: out << "\\nothing"; break;
      case TBehavior: out << "behavior"; break;
      case TComplete: out << "complete"; break;
      case TDisjoint: out << "disjoint"; break;
      case TBehaviors: out << "behaviors"; break;
      case TAt: out << "\\at"; break;
      case TOld: out << "\\old"; break;
      case TResult: out << "\\result"; break;
      case TType: out << "type"; break;
      case TMatch: out << "\\match"; break;
      case TLambda: out << "\\lambda"; break;
      case TSum: out << "\\sum"; break;
      case TProduct: out << "\\product"; break;
      case TLength: out << "\\length"; break;
      case TBlockLength: out << "\\block_length"; break;
      case TOffset: out << "\\offset"; break;
      case TTIso: out << "\\t_iso"; break;
      case TOrigin: out << "\\origin"; break;
      case TObject: out << "\\object"; break;
      case TNull: out << "\\null"; break;
      case TBaseAddr: out << "\\base_addr"; break;
      case TAllocation: out << "\\allocation"; break;
      case TAllocable: out << "\\allocable"; break;
      case TFreeable: out << "\\freeable"; break;
      case TFresh: out << "\\fresh"; break;
      case TSeparated: out << "\\separated"; break;
      case TFrees: out << "\\frees"; break;
      case TAllocates: out << "\\allocates"; break;
      case TAllocationStatus: out << "\\allocation_status"; break;
      case TLStatic: out << "\\static"; break;
      case TLRegister: out << "\\register"; break;
      case TLAutomatic: out << "\\automatic"; break;
      case TLDynamic: out << "\\dynamic"; break;
      case TLUnallocated: out << "\\unallocated"; break;
      case TInitialized: out << "\\initialized"; break;
      case TSpecified: out << "\\specified"; break;
      case TWith: out << "\\with"; break;
      case TNumOf: out << "\\numof"; break;
      case TLUnion: out << "\\union"; break;
      case TLInter: out << "\\inter"; break;
      case TSubset: out << "\\subset"; break;
      case TEmpty: out << "\\empty"; break;
      case TValid: out << "\\valid"; break;
      case TValidRead: out << "\\validread"; break;
      case TValidIndex: out << "\\validindex"; break;
      case TValidRange: out << "\\validrange"; break;
      case TInductive: out << "inductive"; break;
      case TLoop: out << "loop"; break;
      case TVariant: out << "variant"; break;
      case TInvariant: out << "invariant"; break;
      case TSet: out << "\\set"; break;
      case TExits: out << "exits"; break;
      case TBreaks: out << "breaks"; break;
      case TContinues: out << "continues"; break;
      case TReturns: out << "returns"; break;
      case TExitStatus: out << "\\exitstatus"; break;
      case TFrom: out << "\\from"; break;
      case TGlobal: out << "global"; break;
      case TWeak: out << "\\weak"; break;
      case TStrong: out << "\\strong"; break;
      case TGhost: out << "\\ghost"; break;
        //      case TUp: out << "\\Up"; break;
        //      case TDown: out << "\\Down"; break;
        //      case TNearestAway: out << "\\NearestAway"; break;
        //      case TNearestEven: out << "\\NearestEven"; break;
        //      case TRoundFloat: out << "\\round_float"; break;
        //      case TRoundDouble: out << "\\round_double"; break;
        //      case TIsFinite: out << "\\is_finite"; break;
        //      case TIsNaN: out << "\\is_NaN"; break;
        //      case TMin: out << "\\min"; break;
        //      case TMax: out << "\\max"; break;
        //      case TAbs: out << "\\abs"; break;
        //      case TSqrt: out << "\\sqrt"; break;
        //      case TPow: out << "\\pow"; break;
        //      case TCeil: out << "\\ceil"; break;
        //      case TFloor: out << "\\floor"; break;
        //      case TExp: out << "\\exp"; break;
        //      case TLog: out << "\\log"; break;
        //      case TLog10: out << "\\log10"; break;
        //      case TCos: out << "\\cos"; break;
        //      case TSin: out << "\\sin"; break;
        //      case TTan: out << "\\tan"; break;
        //      case TCosh: out << "\\cosh"; break;
        //      case TSinh: out << "\\sinh"; break;
        //      case TTanh: out << "\\tanh"; break;
        //      case TAcos: out << "\\acos"; break;
        //      case TAsin: out << "\\asin"; break;
        //      case TAtan: out << "\\atan"; break;
        //      case TAtan2: out << "\\atan2"; break;
        //      case THypot: out << "\\hypot"; break;
        //      case TExact: out << "\\exact"; break;
        //      case TRoundError: out << "\\round_error"; break;
    };
  };
}

void
LiteralToken::_write(std::ostream& out, PersistentFormat& format) const {
  inherited::_write(out, format);
  if (format.isPretty()) {
    out << " -> ";
    switch (queryOwnField()) {
      case TInteger: out << "int"; break;
      case TCharacter: out << "char"; break;
      case TFloating: out << "double"; break;
      case TString: out << "string"; break;
    };
  };
}

void
IntegerLiteralToken::_write(std::ostream& out, PersistentFormat& format) const {
  inherited::_write(out, format);
  if (!format.isPretty()) { out << _value; }
  else out << " -> " << _value;
}

void
CharacterLiteralToken::_write(std::ostream& out, PersistentFormat& format)
    const {
  inherited::_write(out, format);
  switch (queryOwnField()) {
    case TChar:
      { if (!format.isPretty())
          out.put(_value.ch);
        else
          out << " -> " << _value.ch;
      };
      break;
    case TWideChar:
      { if (!format.isPretty())
          out.write((const char*) &_value.wch, sizeof(wchar_t));
        else
          out << " -> " << _value.wch;
      };
      break;
  };
}

void
FloatingLiteralToken::_write(std::ostream& out, PersistentFormat& format) const
{ inherited::_write(out, format);
  if (!format.isPretty()) out << _real;
  else out << " -> " << _real;
}

void
StringLiteralToken::_write(std::ostream& out, PersistentFormat& format) const {
  inherited::_write(out, format);
  if (!format.isPretty()) {
    unsigned size = _content.length();
    out.write((const char*) &size, sizeof(unsigned));
    out.write(_content.c_str(), size);
  }
  else
    out << " -> " << _content;
}

void
OperatorPunctuatorToken::writeSign(std::ostream& out) const {
  switch (queryOwnField()) {
    case TOpenBrace: out << '{'; break;
    case TCloseBrace: out << '}'; break;
    case TOpenBracket: out << '['; break;
    case TCloseBracket: out << ']'; break;
    case TOpenParen: out << '('; break;
    case TCloseParen: out << ')'; break;
    case TSemiColon: out << ';'; break;
    case TColon: out << ':'; break;
    case TEllipsis: out << "..."; break;
    case TQuery: out << '?'; break;
    case TColonColon: out << "::"; break;
    case TPunct: out << '.'; break;
    case TPunctStar: out << ".*"; break;
    case TPlus: out << '+'; break;
    case TMinus: out << '-'; break;
    case TStar: out << '*'; break;
    case TDivide: out << '/'; break;
    case TProcent: out << '%'; break;
    case TBitXor: out << '^'; break;
    case TAmpersand: out << '&'; break;
    case TBitOr: out << '|'; break;
    case TTilda: out << '~'; break;
    case TNot: out << '!'; break;
    case TAssign: out << '='; break;
    case TLess: out << '<'; break;
    case TGreater: out << '>'; break;
    case TPlusAssign: out << "+="; break;
    case TMinusAssign: out << "-="; break;
    case TMultAssign: out << "*="; break;
    case TDivideAssign: out << "/="; break;
    case TModuloAssign: out << "%="; break;
    case TBitXorAssign: out << "^="; break;
    case TBitAndAssign: out << "&="; break;
    case TBitOrAssign: out << "|="; break;
    case TLeftShift: out << "<<"; break;
    case TRightShift: out << ">>"; break;
    case TLeftShiftAssign: out << "<<="; break;
    case TRightShiftAssign: out << ">>="; break;
    case TEqual: out << "=="; break;
    case TDifferent: out << "!="; break;
    case TLessOrEqual: out << "<="; break;
    case TGreaterOrEqual: out << ">="; break;
    case TLogicalAnd: out << "&&"; break;
    case TLogicalOr: out << "||"; break;
    case TPlusPlus: out << "++"; break;
    case TMinusMinus: out << "--"; break;
    case TComma: out << ','; break;
    case TArrowStar: out << "->*"; break;
    case TArrow: out << "->"; break;
    case TImplies: out << "==>"; break;
    case TEquiv: out << "<==>"; break;
    case TLogicalXor: out << "^^"; break;
    case TBitImplies: out << "-->"; break;
    case TBitEquiv: out << "<-->"; break;
    case TColonGT: out << ":>"; break;
    case TLTColon: out << "<:"; break;
    case TRange: out << ".."; break;
  };
}

void
OperatorPunctuatorToken::_write(std::ostream& out, PersistentFormat& format)
    const {
  inherited::_write(out, format);
  if (format.isPretty()) {
    out << " -> ";
    writeSign(out);
  };
}

void
Token::assumeContent()
{ assert(_type != 0);
  if (!_content.get()) {
    AbstractToken basic(_type);
    basic.clearFullField();
    switch (basic.getType()) {
      case AbstractToken::TKeyword:
        _content.reset(new KeywordToken((const KeywordToken&) basic));
        break;
      case AbstractToken::TOperatorPunctuator:
        _content.reset(new OperatorPunctuatorToken(
              (const OperatorPunctuatorToken&) basic));
        break;
      default:
        _content.reset(new AbstractToken(basic));
    };
  };
}

}} // end of namespace Acsl::DLexer

